==> cego-2.51.1.tar.gz: cego-2.51.1/src/CegoDbThreadPool.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoDbThreadPool.cc  
// -------------------                                                     
// Cego db threadpool class implementation
//                                                         
// Design and Implementation by Bjoern Lemke
//               
// (C)opyright 2000-2025 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoDbThreadPool
// 
// Description: Managing the database threads
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////

// LFC INCLUDES
#include <lfcbase/Exception.h>
#include <lfcbase/NetHandler.h>
#include <lfcbase/Net.h>
#include <lfcbase/Sleeper.h>
#include <lfcbase/ThreadLock.h>
#include <lfcbase/Sleeper.h>
#include <lfcbase/NanoTimer.h>
#include <lfcxml/XMLSuite.h>

// CEGO INCLUDES
#include "CegoXMLdef.h"
#include "CegoDbThreadPool.h"
#include "CegoDefs.h"

// POSIX INCLUDES
#include <string.h>
#include <stdlib.h>

static ThreadLock queueLock("DBQUEUE"); 

// db pool thread array is used to synchronize compiled objects use ( procedures and views ) 
static ThreadLock** thrLockArray; 
extern bool __lockStatOn;

CegoDbThreadPool::CegoDbThreadPool() : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY);
}

CegoDbThreadPool::CegoDbThreadPool(int poolLimit, CegoDatabaseManager *pDBMng, CegoDbHandler::ProtocolType protType)  : Thread()
{
    _poolLimit = poolLimit;
    _samplePos = 0;
    _pDBMng = pDBMng;
    _protType = protType;
    _modId = _pDBMng->getModId("CegoDbThreadPool");
}

CegoDbThreadPool::~CegoDbThreadPool()  
{
    _terminated=true;

    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->abortSession();	
	i++;
    }

    _joined=false;
    int count=0;
    while ( _joined == false && count < POOL_TERMWAIT )
    {
	Sleeper s;
	s.secSleep(1);
	count++;
    }
    
    if ( _joined )
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads terminated"));
	join(getTid());
    }
    else
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Canceling hanging db sessions ..."));
	cancel();
    }
    
    for ( int i = 0; i < _poolLimit; i++ )
    {
	delete _threadList[i];
	delete thrLockArray[i];
    }
    
    delete _numRequest;
    delete _numQueryRequest;
    delete _threadId;
    delete _threadLoad;

    for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
    {
	delete _threadIdle[i];
    }

    delete _threadState;
    delete thrLockArray;
}

void CegoDbThreadPool::P(int i)
{
    thrLockArray[i]->writeLock(DBM_LOCKTIMEOUT);
}

void CegoDbThreadPool::V(int i)
{
    thrLockArray[i]->unlock();
}

void  CegoDbThreadPool::syncToReady()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, __lockStatOn);
    thrLockArray = new ThreadLock*[_poolLimit];

    for (int i = 0; i<_poolLimit; i++)
    {
        thrLockArray[i] = new ThreadLock(Chain("DBTHRLCK") + Chain(i));
        thrLockArray[i]->init(LCKMNG_LOCKWAITDELAY, __lockStatOn);
    }
    
    _pDBMng->getDataPort(_dataPortNo);
    _pDBMng->getDBHost(_dataHostName);

    _maxSendLen = _pDBMng->getMaxSendLen();
    
    _threadId = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _threadLoad = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _numRequest = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _numQueryRequest = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
    {
	_threadIdle[i] = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    }
    _threadState = (ThreadState*)malloc(_poolLimit * sizeof(ThreadState));
    _threadList = (CegoDbThread**)malloc(_poolLimit * sizeof(CegoDbThread*));

    _terminated=false;

    int i=0;
    while ( i < _poolLimit )
    {
	_threadState[i] = STARTING;
	_threadList[i]= new CegoDbThread(this, _pDBMng, _protType);
	_numRequest[i]=0;
	_numQueryRequest[i]=0;

	_threadId[i]=i;
	_threadLoad[i]=0;
	for ( int j = 0; j < THRMNG_NUMLOADSAMPLE; j++ )
	{
	    _threadIdle[j][i]=0;
	}
	_threadList[i]->start(&_threadId[i]);	
	i++;
    }
    
    int numReady=0;
    while ( numReady < _poolLimit )
    {	
	numReady=0;
	int i=0;
	while ( i < _poolLimit )
	{
	    if ( _threadState[i] == READY )
		numReady++;
	    i++;
	}

	Sleeper s;
	s.milliSleep(1);
    }
    _pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads ready"));
}

void  CegoDbThreadPool::loadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->addReloadRequest(tabSetId);
	i++;
    }
}

void  CegoDbThreadPool::unloadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->unloadObjects(tabSetId);
	i++;
    }
}

void CegoDbThreadPool::invalidateObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {	
	try
	{
	    _threadList[i]->invalidateObject(tabSetId, objName, type);
	}
	catch ( Exception e )
	{
	    Chain msg;
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Ignoring error : ") + msg);
	}
	i++;
    }    
}

int CegoDbThreadPool::getPoolLimit() const
{
    return _poolLimit;
}

void CegoDbThreadPool::getThreadInfo(int i, unsigned long long& numRequest, unsigned long long& numQueryRequest, unsigned long long& threadLoad, unsigned long long& numAllocated, CegoDbThreadPool::ThreadState& state)
{
    state = _threadState[i];
    numRequest = _numRequest[i];
    numQueryRequest = _numQueryRequest[i];
    threadLoad = _threadLoad[i];
    numAllocated = _threadList[i]->allocatedSortArea();
}

void CegoDbThreadPool::getThreadHistory(int i, ListT<CegoDbThread::QueryEntry>& queryList)
{
    if ( i < _poolLimit )
	queryList = _threadList[i]->getLastQuery();
}

void CegoDbThreadPool::getQueryHistoryLast(int numLast, ListT<CegoDbThread::QueryEntry>& queryList)
{    
    TreeT<CegoDbThread::QueryEntry> queryTree;
    
    int i=0;
    while ( i < _poolLimit )
    {
	ListT<CegoDbThread::QueryEntry> threadQueryList;
	threadQueryList = _threadList[i]->getLastQuery();

	CegoDbThread::QueryEntry *pEntry =  threadQueryList.First();
	while ( pEntry )
	{
	    pEntry->setRevLastOrder();
	    queryTree.Insert(*pEntry);	    
	    pEntry =  threadQueryList.Next();
	}
	
	i++;
    }

    i=0;
    CegoDbThread::QueryEntry *pEntry =  queryTree.First();
    while ( pEntry && i < numLast )
    {
	queryList.Insert(*pEntry);
	pEntry =  queryTree.Next();
	i++;
    }
}

void CegoDbThreadPool::getQueryHistoryCost(int numLast, ListT<CegoDbThread::QueryEntry>& queryList)
{
    TreeT<CegoDbThread::QueryEntry> queryTree;
    
    int i=0;
    while ( i < _poolLimit )
    {
	ListT<CegoDbThread::QueryEntry> threadQueryList;
	threadQueryList = _threadList[i]->getCostQuery();

	CegoDbThread::QueryEntry *pEntry =  threadQueryList.First();
	while ( pEntry )
	{
	    pEntry->setRevCostOrder();
	    queryTree.Insert(*pEntry);	    
	    pEntry =  threadQueryList.Next();
	}	
	i++;
    }

    i=0;
    CegoDbThread::QueryEntry *pEntry =  queryTree.First();
    while ( pEntry && i < numLast )
    {
	queryList.Insert(*pEntry);
	pEntry =  queryTree.Next();
	i++;
    }
}

void CegoDbThreadPool::setState(int i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void CegoDbThreadPool::incNumRequest(int i)
{
    _numRequest[i]++;
}

void CegoDbThreadPool::incNumQueryRequest(int i)
{
    _numQueryRequest[i]++;
}

void CegoDbThreadPool::addThreadIdle(int i, unsigned long long usec)
{    
    _threadIdle[_samplePos][i] += usec;
}

void CegoDbThreadPool::setTid(int i, unsigned long long tid)
{
    _threadId[i] = tid;
}

void CegoDbThreadPool::setThreadState(int i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void  CegoDbThreadPool::abortThread(int i)
{
    _threadList[i]->abortSession();
}

void* CegoDbThreadPool::job(void* arg)
{
    try
    {
	NanoTimer tim;
	
	Net net( NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN, _maxSendLen );

	if ( _dataHostName != Chain() )
	    net.serve(_dataHostName, _dataPortNo);
	else
	    net.serve6(_dataPortNo);

	unsigned long long usedTime[THRMNG_NUMLOADSAMPLE];
	for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
	{
	    usedTime[i] = 0;
	}

	int selectTimeout = _pDBMng->getSelectTimeout();
	int queueDelay = _pDBMng->getQueueDelay();
	
	while ( ! _terminated ) 
	{
	    usedTime[_samplePos] = 0;
	    
	    tim.reset();
	    tim.start();

	    // take the mutex to avoid cpu load during wait position
	
	    bool noReq=false;
	    lockQueue();
	    if ( _requestQueue.Size() == 0  )  noReq=true;
	    unlockQueue();

	    /*
	      locking the queue to avoid cpu load for idle db threads
	      This leads to fuzzy db thread load statistics.
	      Instead we reduce cpu load with the QUEUEDELAY parameter

	      if ( noReq )
	      lockQueue();
	    */
	    
	    NetHandler* pHandle = 0;

	    try
	    {
		pHandle = net.nextRequest( selectTimeout );	
	    }
	    catch ( Exception e )
	    {
		Chain msg;
		e.pop(msg);
		_pDBMng->log(_modId, Logger::LOGERR, Chain("Error on incoming connection : ") + msg);
	    }
	    
	    if ( noReq )
	    {
		Sleeper s;
		s.microSleep(queueDelay);
	    }
	    
	    if ( pHandle )
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Connection request from <") + pHandle->getSource() + Chain(">"));		
		lockQueue();

		if ( _requestQueue.Size() < NETMNG_MAXQUEUELEN )
		{
		    _requestQueue.Insert(pHandle);
		}
		else
		{
		    delete pHandle;
		    _pDBMng->log(_modId, Logger::NOTICE, Chain("Rejected incoming request since connection queue is full ( ") +  Chain(NETMNG_MAXQUEUELEN) + Chain(" max )"));
		}
		
		unlockQueue();
	    }

	    tim.stop();

	    usedTime[_samplePos] += tim.getSum();
	    
	    tim.reset();
	    tim.start();
	   
	    // calculate load
	    int i=0;
	    while ( i < _poolLimit )
	    {
		unsigned long long ut=0;
		unsigned long long ti=0;
		for ( int j=0; j< THRMNG_NUMLOADSAMPLE; j++ )
		{
		    ut+=usedTime[j];
		    ti+=_threadIdle[j][i];
		}
		
		unsigned long long l = 0;
		if ( ut > 0 && ut > ti )
		    l = (unsigned long long)100 - ( (unsigned long long)100 * ti ) / ut;
				
		_threadLoad[i] = l > 0 ? l : 0; 

		i++;
	    }

	    _samplePos = ( _samplePos + 1 ) % THRMNG_NUMLOADSAMPLE;
	   
	    for ( int i = 0; i < _poolLimit; i++ )
	    {
		_threadIdle[_samplePos][i]=0;
	    }	    
	}
	int i=0;
	while ( i < _poolLimit )
	{
#ifdef CGDEBUG	
	    _pDBMng->log(_modId, Logger::DEBUG, Chain("Waiting for db thread with tid ") + Chain(_threadList[i]->getTid()) + Chain(" to terminate"));
#endif
	    join(_threadList[i]->getTid());
	    i++;
	}
	_joined=true;
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);	
	_pDBMng->log(_modId, Logger::LOGERR, Chain("Db Thread start failed : ") +  msg);	
	_terminated=true;
    }

    // waiting for db threads to terminate    
    return 0;
}

bool CegoDbThreadPool::isBusy()
{
    int i=0;
    while ( i < _poolLimit )
    {
	if ( _threadState[i] == BUSY )
	    return true;
	i++;
    }
    return false;
}

bool CegoDbThreadPool::isTerminated()
{
    return _terminated;
}

NetHandler* CegoDbThreadPool::nextRequest()
{
    lockQueue();

    NetHandler **pRequest = _requestQueue.First();

    if ( pRequest )
    {
	NetHandler *pN = *pRequest;
	_requestQueue.Remove(*pRequest);
	unlockQueue();

	return pN; 
    }
    else
    {
	unlockQueue();
	return 0;
    }    
}

void CegoDbThreadPool::lockQueue()
{
    queueLock.writeLock();
}

void CegoDbThreadPool::unlockQueue()
{
    queueLock.unlock();
}

void CegoDbThreadPool::getLockStat(Chain& lockName, int& lockCount, unsigned long long &numRdLock, unsigned long long &numWrLock, unsigned long long &sumRdDelay, unsigned long long &sumWrDelay)
{
    lockName = queueLock.getId();
    lockCount = queueLock.numLockTry();

    numRdLock = queueLock.numReadLock();
    numWrLock = queueLock.numWriteLock();

    sumRdDelay = 0;
    sumWrDelay = 0;

    if ( queueLock.numReadLock() > 0 )
	sumRdDelay = queueLock.sumReadDelay() / LCKMNG_DELRES;
    if ( queueLock.numWriteLock() > 0 )
	sumWrDelay = queueLock.sumWriteDelay() / LCKMNG_DELRES;
}

void CegoDbThreadPool::getAggThreadLockStat(Chain& lockGroup, int& numLock, int& lockCount, unsigned long long &numRdLock, unsigned long long &numWrLock, unsigned long long &sumRdDelay, unsigned long long &sumWrDelay)
{
    lockGroup = Chain("DBTHRLCK");
    
    lockCount = 0;
    numRdLock = 0;
    numWrLock = 0;
    sumRdDelay = 0;
    sumWrDelay = 0;
    numLock = 0;

    for ( int i = 0; i<_poolLimit; i++)
    {	
	numLock++;
	lockCount += thrLockArray[i]->numLockTry();
	numRdLock += thrLockArray[i]->numReadLock();
	numWrLock += thrLockArray[i]->numWriteLock();
	    
	if ( thrLockArray[i]->numReadLock() > 0 )
	    sumRdDelay += thrLockArray[i]->sumReadDelay() / LCKMNG_DELRES;   
	if ( thrLockArray[i]->numWriteLock() > 0 )
	    sumWrDelay += thrLockArray[i]->sumWriteDelay() / LCKMNG_DELRES;
    }
}

==> cego-2.52.42.tar.gz: cego-2.52.42/src/CegoDbThreadPool.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoDbThreadPool.cc  
// -------------------                                                     
// Cego db threadpool class implementation
//                                                         
// Design and Implementation by Bjoern Lemke
//               
// (C)opyright 2000-2025 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoDbThreadPool
// 
// Description: Managing the database threads
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////

// LFC INCLUDES
#include <lfcbase/Exception.h>
#include <lfcbase/NetHandler.h>
#include <lfcbase/Net.h>
#include <lfcbase/Sleeper.h>
#include <lfcbase/ThreadLock.h>
#include <lfcbase/Sleeper.h>
#include <lfcbase/NanoTimer.h>
#include <lfcxml/XMLSuite.h>

// CEGO INCLUDES
#include "CegoXMLdef.h"
#include "CegoDbThreadPool.h"
#include "CegoDefs.h"

// POSIX INCLUDES
#include <string.h>
#include <stdlib.h>

static ThreadLock queueLock("DBQUEUE"); 

// db pool thread array is used to synchronize compiled objects use ( procedures and views ) 
static ThreadLock** thrLockArray; 
extern bool __lockStatOn;

CegoDbThreadPool::CegoDbThreadPool() : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY);
}

CegoDbThreadPool::CegoDbThreadPool(int poolLimit, CegoDatabaseManager *pDBMng, CegoDbHandler::ProtocolType protType)  : Thread()
{
    _poolLimit = poolLimit;
    _samplePos = 0;
    _pDBMng = pDBMng;
    _protType = protType;
    _modId = _pDBMng->getModId("CegoDbThreadPool");
}

CegoDbThreadPool::~CegoDbThreadPool()  
{
    _terminated=true;

    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->abortSession();	
	i++;
    }

    _joined=false;
    int count=0;
    while ( _joined == false && count < POOL_TERMWAIT )
    {
	Sleeper s;
	s.secSleep(1);
	count++;
    }
    
    if ( _joined )
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads terminated"));
	join(getTid());
    }
    else
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Canceling hanging db sessions ..."));
	cancel();
    }
    
    for ( int i = 0; i < _poolLimit; i++ )
    {
	delete _threadList[i];
	delete thrLockArray[i];
    }
    
    delete _numRequest;
    delete _numQueryRequest;
    delete _threadId;
    delete _threadLoad;

    for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
    {
	delete _threadIdle[i];
    }

    delete _threadState;
    delete thrLockArray;
}

void CegoDbThreadPool::P(int i)
{
    thrLockArray[i]->writeLock(DBM_LOCKTIMEOUT);
}

void CegoDbThreadPool::V(int i)
{
    thrLockArray[i]->unlock();
}

void  CegoDbThreadPool::syncToReady()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, __lockStatOn);
    thrLockArray = new ThreadLock*[_poolLimit];

    for (int i = 0; i<_poolLimit; i++)
    {
        thrLockArray[i] = new ThreadLock(Chain("DBTHRLCK") + Chain(i));
        thrLockArray[i]->init(LCKMNG_LOCKWAITDELAY, __lockStatOn);
    }
    
    _pDBMng->getDataPort(_dataPortNo);
    _pDBMng->getDBHost(_dataHostName);

    _maxSendLen = _pDBMng->getMaxSendLen();
    
    _threadId = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _threadLoad = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _numRequest = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    _numQueryRequest = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
    {
	_threadIdle[i] = (unsigned long long*)malloc(_poolLimit * sizeof(unsigned long long));
    }
    _threadState = (ThreadState*)malloc(_poolLimit * sizeof(ThreadState));
    _threadList = (CegoDbThread**)malloc(_poolLimit * sizeof(CegoDbThread*));

    _terminated=false;

    int i=0;
    while ( i < _poolLimit )
    {
	_threadState[i] = STARTING;
	_threadList[i]= new CegoDbThread(this, _pDBMng, _protType);
	_numRequest[i]=0;
	_numQueryRequest[i]=0;

	_threadId[i]=i;
	_threadLoad[i]=0;
	for ( int j = 0; j < THRMNG_NUMLOADSAMPLE; j++ )
	{
	    _threadIdle[j][i]=0;
	}
	_threadList[i]->start(&_threadId[i]);	
	i++;
    }
    
    int numReady=0;
    while ( numReady < _poolLimit )
    {	
	numReady=0;
	int i=0;
	while ( i < _poolLimit )
	{
	    if ( _threadState[i] == READY )
		numReady++;
	    i++;
	}

	Sleeper s;
	s.milliSleep(1);
    }
    _pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads ready"));
}

void  CegoDbThreadPool::loadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->addReloadRequest(tabSetId);
	i++;
    }
}

void  CegoDbThreadPool::unloadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->unloadObjects(tabSetId);
	i++;
    }
}

void CegoDbThreadPool::invalidateObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {	
	try
	{
	    _threadList[i]->invalidateObject(tabSetId, objName, type);
	}
	catch ( Exception e )
	{
	    Chain msg;
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Ignoring error : ") + msg);
	}
	i++;
    }    
}

int CegoDbThreadPool::getPoolLimit() const
{
    return _poolLimit;
}

void CegoDbThreadPool::getThreadInfo(int i, unsigned long long& numRequest, unsigned long long& numQueryRequest, unsigned long long& threadLoad, unsigned long long& numAllocated, CegoDbThreadPool::ThreadState& state, Chain& lastAction)
{
    state = _threadState[i];
    numRequest = _numRequest[i];
    numQueryRequest = _numQueryRequest[i];
    threadLoad = _threadLoad[i];
    numAllocated = _threadList[i]->allocatedSortArea();
    lastAction = _threadList[i]->lastAction();
}

void CegoDbThreadPool::getThreadHistory(int i, ListT<CegoDbThread::QueryEntry>& queryList)
{
    if ( i < _poolLimit )
	queryList = _threadList[i]->getLastQuery();
}

void CegoDbThreadPool::getQueryHistoryLast(int numLast, ListT<CegoDbThread::QueryEntry>& queryList)
{    
    TreeT<CegoDbThread::QueryEntry> queryTree;
    
    int i=0;
    while ( i < _poolLimit )
    {
	ListT<CegoDbThread::QueryEntry> threadQueryList;
	threadQueryList = _threadList[i]->getLastQuery();

	CegoDbThread::QueryEntry *pEntry =  threadQueryList.First();
	while ( pEntry )
	{
	    pEntry->setRevLastOrder();
	    queryTree.Insert(*pEntry);	    
	    pEntry =  threadQueryList.Next();
	}
	
	i++;
    }

    i=0;
    CegoDbThread::QueryEntry *pEntry =  queryTree.First();
    while ( pEntry && i < numLast )
    {
	queryList.Insert(*pEntry);
	pEntry =  queryTree.Next();
	i++;
    }
}

void CegoDbThreadPool::getQueryHistoryCost(int numLast, ListT<CegoDbThread::QueryEntry>& queryList)
{
    TreeT<CegoDbThread::QueryEntry> queryTree;
    
    int i=0;
    while ( i < _poolLimit )
    {
	ListT<CegoDbThread::QueryEntry> threadQueryList;
	threadQueryList = _threadList[i]->getCostQuery();

	CegoDbThread::QueryEntry *pEntry =  threadQueryList.First();
	while ( pEntry )
	{
	    pEntry->setRevCostOrder();
	    queryTree.Insert(*pEntry);	    
	    pEntry =  threadQueryList.Next();
	}	
	i++;
    }

    i=0;
    CegoDbThread::QueryEntry *pEntry =  queryTree.First();
    while ( pEntry && i < numLast )
    {
	queryList.Insert(*pEntry);
	pEntry =  queryTree.Next();
	i++;
    }
}

void CegoDbThreadPool::setState(int i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void CegoDbThreadPool::incNumRequest(int i)
{
    _numRequest[i]++;
}

void CegoDbThreadPool::incNumQueryRequest(int i)
{
    _numQueryRequest[i]++;
}

void CegoDbThreadPool::addThreadIdle(int i, unsigned long long usec)
{    
    _threadIdle[_samplePos][i] += usec;
}

void CegoDbThreadPool::setTid(int i, unsigned long long tid)
{
    _threadId[i] = tid;
}

void CegoDbThreadPool::setThreadState(int i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void  CegoDbThreadPool::abortThread(int i)
{
    _threadList[i]->abortSession();
}

void* CegoDbThreadPool::job(void* arg)
{
    try
    {
	NanoTimer tim;
	
	Net net( NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN, _maxSendLen );

	if ( _dataHostName != Chain() )
	    net.serve(_dataHostName, _dataPortNo);
	else
	    net.serve6(_dataPortNo);

	unsigned long long usedTime[THRMNG_NUMLOADSAMPLE];
	for ( int i = 0; i < THRMNG_NUMLOADSAMPLE; i++ )
	{
	    usedTime[i] = 0;
	}

	int selectTimeout = _pDBMng->getSelectTimeout();
	int queueDelay = _pDBMng->getQueueDelay();
	
	while ( ! _terminated ) 
	{
	    usedTime[_samplePos] = 0;
	    
	    tim.reset();
	    tim.start();

	    // take the mutex to avoid cpu load during wait position
	
	    bool noReq=false;
	    lockQueue();
	    if ( _requestQueue.Size() == 0  )  noReq=true;
	    unlockQueue();

	    /*
	      locking the queue to avoid cpu load for idle db threads
	      This leads to fuzzy db thread load statistics.
	      Instead we reduce cpu load with the QUEUEDELAY parameter

	      if ( noReq )
	      lockQueue();
	    */
	    
	    NetHandler* pHandle = 0;

	    try
	    {
		pHandle = net.nextRequest( selectTimeout );	
	    }
	    catch ( Exception e )
	    {
		Chain msg;
		e.pop(msg);
		_pDBMng->log(_modId, Logger::LOGERR, Chain("Error on incoming connection : ") + msg);
	    }
	    
	    if ( noReq )
	    {
		Sleeper s;
		s.microSleep(queueDelay);
	    }
	    
	    if ( pHandle )
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Connection request from <") + pHandle->getSource() + Chain(">"));		
		lockQueue();

		if ( _requestQueue.Size() < NETMNG_MAXQUEUELEN )
		{
		    _requestQueue.Insert(pHandle);
		}
		else
		{
		    delete pHandle;
		    _pDBMng->log(_modId, Logger::NOTICE, Chain("Rejected incoming request since connection queue is full ( ") +  Chain(NETMNG_MAXQUEUELEN) + Chain(" max )"));
		}
		
		unlockQueue();
	    }

	    tim.stop();

	    usedTime[_samplePos] += tim.getSum();
	    
	    tim.reset();
	    tim.start();
	   
	    // calculate load
	    int i=0;
	    while ( i < _poolLimit )
	    {
		unsigned long long ut=0;
		unsigned long long ti=0;
		for ( int j=0; j< THRMNG_NUMLOADSAMPLE; j++ )
		{
		    ut+=usedTime[j];
		    ti+=_threadIdle[j][i];
		}
		
		unsigned long long l = 0;
		if ( ut > 0 && ut > ti )
		    l = (unsigned long long)100 - ( (unsigned long long)100 * ti ) / ut;
				
		_threadLoad[i] = l > 0 ? l : 0; 

		i++;
	    }

	    _samplePos = ( _samplePos + 1 ) % THRMNG_NUMLOADSAMPLE;
	   
	    for ( int i = 0; i < _poolLimit; i++ )
	    {
		_threadIdle[_samplePos][i]=0;
	    }	    
	}
	int i=0;
	while ( i < _poolLimit )
	{
#ifdef CGDEBUG	
	    _pDBMng->log(_modId, Logger::DEBUG, Chain("Waiting for db thread with tid ") + Chain(_threadList[i]->getTid()) + Chain(" to terminate"));
#endif
	    join(_threadList[i]->getTid());
	    i++;
	}
	_joined=true;
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);	
	_pDBMng->log(_modId, Logger::LOGERR, Chain("Db Thread start failed : ") +  msg);	
	_terminated=true;
    }

    // waiting for db threads to terminate    
    return 0;
}

bool CegoDbThreadPool::isBusy()
{
    int i=0;
    while ( i < _poolLimit )
    {
	if ( _threadState[i] == BUSY )
	    return true;
	i++;
    }
    return false;
}

bool CegoDbThreadPool::isTerminated()
{
    return _terminated;
}

NetHandler* CegoDbThreadPool::nextRequest()
{
    lockQueue();

    NetHandler **pRequest = _requestQueue.First();

    if ( pRequest )
    {
	NetHandler *pN = *pRequest;
	_requestQueue.Remove(*pRequest);
	unlockQueue();

	return pN; 
    }
    else
    {
	unlockQueue();
	return 0;
    }    
}

void CegoDbThreadPool::lockQueue()
{
    queueLock.writeLock();
}

void CegoDbThreadPool::unlockQueue()
{
    queueLock.unlock();
}

void CegoDbThreadPool::getLockStat(Chain& lockName, int& lockCount, unsigned long long &numRdLock, unsigned long long &numWrLock, unsigned long long &sumRdDelay, unsigned long long &sumWrDelay)
{
    lockName = queueLock.getId();
    lockCount = queueLock.numLockTry();

    numRdLock = queueLock.numReadLock();
    numWrLock = queueLock.numWriteLock();

    sumRdDelay = 0;
    sumWrDelay = 0;

    if ( queueLock.numReadLock() > 0 )
	sumRdDelay = queueLock.sumReadDelay() / LCKMNG_DELRES;
    if ( queueLock.numWriteLock() > 0 )
	sumWrDelay = queueLock.sumWriteDelay() / LCKMNG_DELRES;
}

void CegoDbThreadPool::getAggThreadLockStat(Chain& lockGroup, int& numLock, int& lockCount, unsigned long long &numRdLock, unsigned long long &numWrLock, unsigned long long &sumRdDelay, unsigned long long &sumWrDelay)
{
    lockGroup = Chain("DBTHRLCK");
    
    lockCount = 0;
    numRdLock = 0;
    numWrLock = 0;
    sumRdDelay = 0;
    sumWrDelay = 0;
    numLock = 0;

    for ( int i = 0; i<_poolLimit; i++)
    {	
	numLock++;
	lockCount += thrLockArray[i]->numLockTry();
	numRdLock += thrLockArray[i]->numReadLock();
	numWrLock += thrLockArray[i]->numWriteLock();
	    
	if ( thrLockArray[i]->numReadLock() > 0 )
	    sumRdDelay += thrLockArray[i]->sumReadDelay() / LCKMNG_DELRES;   
	if ( thrLockArray[i]->numWriteLock() > 0 )
	    sumWrDelay += thrLockArray[i]->sumWriteDelay() / LCKMNG_DELRES;
    }
}
